# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
from abc import ABC
from abc import abstractmethod
from enum import IntEnum
from enum import ReprEnum
from typing import override


class HardwareTypes(IntEnum):
    """Hardware types for the Macropad."""

    UNKNOWN = 0x00
    FIVE_BUTTON_USB_V1 = 0x02
    FIVE_BUTTON_USB = 0x03
    FIVE_BUTTON_BT = 0x04
    TEN_BUTTON_USB = 0x05
    TEN_BUTTON_BT = 0x06


class HidInCommands(IntEnum):
    """Identifiers for incoming HID messages."""

    VERSION_INFO = 0x99
    STATUS = 0x1
    STATUS_REQUEST = 0x2


class HidOutCommands(IntEnum):
    """Identifiers for outgoing HID messages."""

    SET_LED = 0x1
    PING = 0xF0
    PREPARE_UPDATE = 0xE0
    RESET = 0xE1
    UPDATE_CONFIG = 0xE2


class HidInputMessage:
    MESSAGE_LENGTH = 8

    @staticmethod
    def from_buffer(buffer: bytes):
        match buffer[1]:
            case HidInCommands.VERSION_INFO:
                return VersionInfo(buffer[2:8])
            case HidInCommands.STATUS:
                return Status(buffer[2:8])
            case HidInCommands.STATUS_REQUEST.value:
                return StatusRequest()
        raise NotImplementedError

    def __repr__(self):  # pragma: no cover
        return self.__str__()


class Status(HidInputMessage):
    @classmethod
    def trigger_button(cls, button: int):
        return cls(bytes([button, 1, 0, 0, 1]))

    def __init__(self, buffer: bytes):
        self.buffer = buffer

    def __str__(self):
        return (
            f"Status {{ button: {self.button}, triggered: {self.triggered}, "
            f"longpress: {self.longpressed}, pressed: {self.pressed}, "
            f"released: {self.released} }}"
        )

    @property
    def button(self):
        return self.buffer[0]

    @property
    def triggered(self) -> bool:
        return self.buffer[1] != 0

    @property
    def longpressed(self) -> bool:
        return self.buffer[2] != 0

    @property
    def pressed(self) -> bool:
        return self.buffer[3] != 0

    @property
    def released(self) -> bool:
        return self.buffer[4] != 0


class VersionInfo(HidInputMessage):
    def __init__(self, buffer: bytes):
        self.buffer = buffer

    def __str__(self):
        return f"Version Info: {self.version}, type {self.type.name}"

    @property
    def version(self):
        return f"{self.buffer[0]}.{self.buffer[1]}.{self.buffer[2]}"

    @property
    def type(self):
        return HardwareTypes(self.buffer[3])


class StatusRequest(HidInputMessage):
    def __str__(self):  # pragma: no cover
        return "Status Request"


class HidOutputMessage:
    REPORT_ID = 1
    pass


class HidCommand(HidOutputMessage, ABC):
    REPORT_ID = 1
    _counter = 0

    def __init__(self):
        type(self)._counter = (type(self)._counter + 1) % 256
        self._current_counter = type(self)._counter

    @abstractmethod
    def to_buffer(self) -> bytes:  # pragma: no cover
        raise NotImplementedError


class LedColor(tuple, ReprEnum):
    # The colors are encoded Green, Red, Blue, White
    RED = (0x00, 0x0A, 0x00, 0x00)
    GREEN = (0x0A, 0x00, 0x00, 0x00)
    BLUE = (0x00, 0x00, 0x0A, 0x00)
    WHITE = (0x00, 0x00, 0x00, 0x0A)
    BLACK = (0x00, 0x00, 0x00, 0x00)
    YELLOW = (0x0A, 0x0A, 0x00, 0x00)
    CYAN = (0x0A, 0x00, 0x0A, 0x00)
    MAGENTA = (0x00, 0x0A, 0x0A, 0x00)
    ORANGE = (0x08, 0x0A, 0x00, 0x00)
    PURPLE = (0x00, 0x09, 0x09, 0x00)


class SetLed(HidCommand):
    def __init__(self, id, led_color: LedColor):
        super().__init__()
        self.id = id
        self.color = led_color

    @override
    def to_buffer(self) -> bytes:
        color = self.color.value
        return bytes(
            [
                HidOutCommands.SET_LED,
                self.id,
                color[0],
                color[1],
                color[2],
                color[3],
                0,
                self._current_counter,
            ],
        )

    def __eq__(self, other):
        return self.id == other.id and self.color == other.color

    def __str__(self):
        return f"SetLed {{ id: {self.id}, color: {self.color.name} }}"


class UpdateConfig(HidCommand):
    def __init__(self):
        super().__init__()
        self._activate_debug = 0
        self._activate_filesystem = 0

    def activate_serial_console(self, activate: bool):
        self._activate_debug = 2 if activate else 1

    def activate_filesystem(self, activate: bool):
        self._activate_filesystem = 2 if activate else 1

    @override
    def to_buffer(self) -> bytes:
        return bytes(
            [
                HidOutCommands.UPDATE_CONFIG,
                self._activate_debug,
                self._activate_filesystem,
                0,
                0,
                0,
                0,
                self._current_counter,
            ],
        )

    def __str__(self):
        return f"UpdateConfig {{ debug: {self._activate_debug}, filesystem: {self._activate_filesystem} }}"


class SimpleHidCommand(HidCommand):
    def __init__(self, command: HidOutCommands):
        super().__init__()
        self.command = command

    @override
    def to_buffer(self) -> bytes:
        return bytes([int(self.command), 0, 0, 0, 0, 0, 0, self._current_counter])

    def __str__(self):  # pragma: no cover
        return f"{self.command.name}"

    def __repr__(self):  # pragma: no cover
        return self.__str__()


class Ping(SimpleHidCommand):
    def __init__(self):
        super().__init__(HidOutCommands.PING)


class PrepareUpdate(SimpleHidCommand):
    def __init__(self):
        super().__init__(HidOutCommands.PREPARE_UPDATE)


class Reset(SimpleHidCommand):
    def __init__(self):
        super().__init__(HidOutCommands.RESET)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import logging

import pytest


@pytest.fixture(scope="session", autouse=True)
def loggingconfig():
    print("Setting up logging configuration")
    logging.basicConfig(level=logging.DEBUG)
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    for logger in loggers:
        logger.setLevel(logging.DEBUG)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import os
from pathlib import Path
from unittest.mock import mock_open
from unittest.mock import patch

import yaml
from mutenix.config import ActionEnum
from mutenix.config import button_action_details_descriminator
from mutenix.config import ButtonAction
from mutenix.config import Config
from mutenix.config import CONFIG_FILENAME
from mutenix.config import create_default_config
from mutenix.config import find_config_file
from mutenix.config import KeyPress
from mutenix.config import load_config
from mutenix.config import MouseActionClick
from mutenix.config import save_config
from mutenix.config import WebhookAction
from mutenix.teams_messages import MeetingAction


def test_find_config_file_default_location():
    with patch("pathlib.Path.exists", return_value=True):
        config_path = find_config_file()
        assert config_path == Path(CONFIG_FILENAME)


def test_find_config_file_not_found():
    with patch("pathlib.Path.exists", return_value=False):
        config_path = find_config_file()
        assert config_path == Path(CONFIG_FILENAME)


def test_load_config_default():
    with patch("pathlib.Path.exists", return_value=False):
        with patch("builtins.open", mock_open(read_data="")):
            with patch(
                "mutenix.config.create_default_config",
            ) as mock_create_default_config:
                default_config = create_default_config()
                default_config.file_path = str(Path(CONFIG_FILENAME))
                mock_create_default_config.return_value = default_config
                config = load_config()
                assert config == mock_create_default_config.return_value


def test_load_config_file_not_found():
    with patch("pathlib.Path.exists", return_value=False):
        with patch("builtins.open", mock_open(read_data="")):
            with patch(
                "mutenix.config.create_default_config",
            ) as mock_create_default_config:
                mock_create_default_config.return_value = create_default_config()
                with patch("mutenix.config.save_config") as mock_save_config:
                    config = load_config()
                    assert config == mock_create_default_config.return_value
                    mock_save_config.assert_not_called()


def test_load_config_yaml_error():
    with patch("pathlib.Path.exists", return_value=True):
        with patch("builtins.open", mock_open(read_data="invalid_yaml")):
            with patch("yaml.safe_load", side_effect=yaml.YAMLError):
                with patch(
                    "mutenix.config.create_default_config",
                ) as mock_create_default_config:
                    mock_create_default_config.return_value = create_default_config()
                    with patch("mutenix.config.save_config") as mock_save_config:
                        config = load_config()
                        assert config == mock_create_default_config.return_value
                    mock_save_config.assert_not_called()


def test_load_config_success():
    config_data = {
        "actions": [
            {"button_id": 1, "action": "toggle-mute"},
            {"button_id": 2, "action": "toggle-hand"},
        ],
        "longpress_action": [],
        "teams_token": None,
    }
    with patch("pathlib.Path.exists", return_value=True):
        with patch("builtins.open", mock_open(read_data=yaml.dump(config_data))):
            with patch("yaml.safe_load", return_value=config_data):
                config = load_config()
                assert config == Config(
                    **config_data,
                )


def test_config_initialization():
    config_data = {
        "actions": [
            {"button_id": 1, "action": "toggle-mute"},
            {"button_id": 2, "action": "toggle-hand"},
        ],
        "longpress_action": [],
        "leds": [],
        "teams_token": None,
        "file_path": None,
        "virtual_keypad": {"bind_address": "127.0.0.1", "bind_port": 12909},
        "auto_update": True,
        "device_identifications": [],
    }
    config = Config(**config_data)
    assert config.actions[0].button_id == 1
    assert config.actions[0].action == "toggle-mute"
    assert config.virtual_keypad.bind_address == "127.0.0.1"
    assert config.virtual_keypad.bind_port == 12909
    assert config.auto_update is True


def test_load_config_with_valid_file():
    config_data = {
        "actions": [
            {"button_id": 1, "action": "toggle-mute"},
            {"button_id": 2, "action": "toggle-hand"},
        ],
        "longpress_action": [],
        "leds": [],
        "teams_token": None,
        "file_path": None,
        "virtual_keypad": {"bind_address": "127.0.0.1", "bind_port": 12909},
        "auto_update": True,
        "device_identifications": [],
    }
    with patch("pathlib.Path.exists", return_value=True):
        with patch("builtins.open", mock_open(read_data=yaml.dump(config_data))):
            with patch("yaml.safe_load", return_value=config_data):
                config = load_config()
                assert config.actions[0].button_id == 1
                assert config.actions[0].action == "toggle-mute"
                assert config.virtual_keypad.bind_address == "127.0.0.1"
                assert config.virtual_keypad.bind_port == 12909
                assert config.auto_update is True


def test_load_config_with_invalid_yaml():
    with patch("pathlib.Path.exists", return_value=True):
        with patch("builtins.open", mock_open(read_data="invalid_yaml")):
            with patch("yaml.safe_load", side_effect=yaml.YAMLError):
                with patch(
                    "mutenix.config.create_default_config",
                ) as mock_create_default_config:
                    mock_create_default_config.return_value = create_default_config()
                    with patch("mutenix.config.save_config") as mock_save_config:
                        config = load_config()
                        assert config == mock_create_default_config.return_value
                        mock_save_config.assert_not_called()


def test_save_config():
    config = create_default_config()
    with patch("builtins.open", mock_open()) as mocked_file:
        save_config(config, "test_config.yaml")
        mocked_file.assert_called_once_with("test_config.yaml", "w")
        handle = mocked_file()
        handle.write.assert_called()


def test_button_action_with_none_extra():
    action = ButtonAction(button_id=1, action=MeetingAction.ToggleMute)
    assert action.button_id == 1
    assert action.action == MeetingAction.ToggleMute
    assert action.extra is None


def test_button_action_with_react_extra():
    action = ButtonAction(button_id=1, action=MeetingAction.React, extra="like")
    assert action.button_id == 1
    assert action.action == MeetingAction.React
    assert action.extra == "like"


def test_button_action_with_sequence_extra():
    sequence = [
        MouseActionClick(action="click", button="left").model_dump(),
        MouseActionClick(action="click", button="left").model_dump(),
    ]
    action = ButtonAction(button_id=1, action=ActionEnum.CMD, extra=sequence)
    assert action.button_id == 1
    assert action.action == ActionEnum.CMD
    assert len(action.extra) == 2


def test_button_action_with_single_keypress_extra():
    keypress = KeyPress(modifiers=["ctrl"], key="a")
    keypress_extra = keypress.model_dump()
    action = ButtonAction(button_id=1, action=ActionEnum.KEYPRESS, extra=keypress_extra)
    assert action.button_id == 1
    assert action.action == ActionEnum.KEYPRESS
    assert action.extra == keypress


def test_button_action_with_single_mouse_action_extra():
    mouse_action = MouseActionClick(action="click", button="left")
    mouse_action_extra = mouse_action.model_dump()
    action = ButtonAction(
        button_id=1,
        action=ActionEnum.MOUSE,
        extra=mouse_action_extra,
    )
    assert action.button_id == 1
    assert action.action == ActionEnum.MOUSE
    assert action.extra == mouse_action


def test_button_action_with_webhook_extra():
    webhook = WebhookAction(
        method="POST",
        url="http://example.com",
        headers={"Content-Type": "application/json"},
        data={"key": "value"},
    )
    webhook_extra = webhook.model_dump()
    action = ButtonAction(button_id=1, action=ActionEnum.WEBHOOK, extra=webhook_extra)
    assert action.button_id == 1
    assert action.action == ActionEnum.WEBHOOK
    assert action.extra == webhook


def test_button_action_details_descriminator_cmd():
    assert button_action_details_descriminator("some_command") == "cmd"


def test_button_action_details_descriminator_key():
    keypress = {"key": "a", "modifiers": ["ctrl"]}
    assert button_action_details_descriminator(keypress) == "key"


def test_button_action_details_descriminator_mouse():
    mouse_action = {"x": 100, "y": 200, "button": "left"}
    assert button_action_details_descriminator(mouse_action) == "mouse"


def test_button_action_details_descriminator_webhook():
    webhook_action = {"url": "http://example.com", "method": "POST"}
    assert button_action_details_descriminator(webhook_action) == "webhook"


def test_button_action_details_descriminator_empty_dict():
    assert button_action_details_descriminator({}) == ""


def test_button_action_details_descriminator_invalid_type():
    assert button_action_details_descriminator(123) == ""


def test_find_config_file_in_current_directory():
    with patch("pathlib.Path.exists", side_effect=[True, False]):
        config_path = find_config_file()
        assert config_path == Path(CONFIG_FILENAME)


def test_find_config_file_in_home_directory():
    with patch("pathlib.Path.exists", side_effect=[False, True]):
        home_config_path = (
            Path.home() / os.environ.get("XDG_CONFIG_HOME", ".config") / CONFIG_FILENAME
        )
        config_path = find_config_file()
        assert config_path == home_config_path


def test_find_config_file_not_found_anywhere():
    with patch("pathlib.Path.exists", return_value=False):
        config_path = find_config_file()
        assert config_path == Path(CONFIG_FILENAME)


def test_load_config_with_file_path():
    config_data = {
        "actions": [
            {"button_id": 1, "action": "toggle-mute"},
            {"button_id": 2, "action": "toggle-hand"},
        ],
        "longpress_action": [],
        "leds": [],
        "teams_token": None,
        "file_path": None,
        "virtual_keypad": {"bind_address": "127.0.0.1", "bind_port": 12909},
        "auto_update": True,
        "device_identifications": [],
    }
    with patch("pathlib.Path.exists", return_value=True):
        with patch("builtins.open", mock_open(read_data=yaml.dump(config_data))):
            with patch("yaml.safe_load", return_value=config_data):
                config = load_config(Path("custom_config.yaml"))
                assert config.actions[0].button_id == 1
                assert config.actions[0].action == "toggle-mute"
                assert config.virtual_keypad.bind_address == "127.0.0.1"
                assert config.virtual_keypad.bind_port == 12909
                assert config.auto_update is True



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import logging
import os
from enum import Enum
from pathlib import Path
from typing import Annotated
from typing import Any
from typing import Literal
from typing import Union

import pydantic
import yaml
from mutenix.teams_messages import ClientMessageParameterType
from mutenix.teams_messages import MeetingAction
from pydantic import BaseModel
from pydantic import Discriminator
from pydantic import Tag

_logger = logging.getLogger(__name__)

CONFIG_FILENAME = "mutenix.yaml"


class ActionEnum(str, Enum):
    """
    ActionEnum is an enumeration that represents different types of actions that can be performed.
    """

    TEAMS = "teams"
    ACTIVATE_TEAMS = "activate-teams"
    CMD = "cmd"
    WEBHOOK = "webhook"
    KEYPRESS = "key-press"
    MOUSE = "mouse"


class LedStatusSource(str, Enum):
    TEAMS = "teams"
    CMD = "cmd"
    WEBHOOK = "webhook"


class LedColor(str, Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"
    WHITE = "white"
    BLACK = "black"
    YELLOW = "yellow"
    CYAN = "cyan"
    MAGENTA = "magenta"
    ORANGE = "orange"
    PURPLE = "purple"


class TeamsState(str, Enum):
    MUTED = "is-muted"
    HAND_RAISED = "is-hand-raised"
    IN_MEETING = "is-in-meeting"
    RECORDING_ON = "is-recording-on"
    BACKGROUND_BLURRED = "is-background-blurred"
    SHARING = "is-sharing"
    UNREAD_MESSAGES = "has-unread-messages"
    VIDEO_ON = "is-video-on"


class TeamsReact(str, Enum):
    reaction: ClientMessageParameterType


class KeyPress(BaseModel):
    modifiers: list[str] | None = None
    key: str | None = None
    string: str | None = None


class MouseActionPosition(BaseModel):
    x: int
    y: int


class MouseActionMove(MouseActionPosition):
    action: Literal["move", None] = "move"


class MouseActionSetPosition(MouseActionPosition):
    action: Literal["set"] = "set"


class MouseActionClick(BaseModel):
    action: Literal["click"] = "click"
    button: str
    count: int = 1


class MouseActionPress(BaseModel):
    action: Literal["press"] = "press"
    button: str


class MouseActionRelease(BaseModel):
    action: Literal["release"] = "release"
    button: str


MouseMove = Annotated[
    Union[
        Annotated[MouseActionMove, Tag("move")],
        Annotated[MouseActionSetPosition, Tag("set")],
        Annotated[MouseActionClick, Tag("click")],
        Annotated[MouseActionPress, Tag("press")],
        Annotated[MouseActionRelease, Tag("release")],
    ],
    Discriminator(lambda v: v["action"] if "action" in v else "move"),
]


class WebhookAction(BaseModel):
    method: str = "GET"
    url: str
    headers: dict[str, str] = {}
    data: dict[str, Any] | None = None


def button_action_details_descriminator(v: Any) -> str:
    if isinstance(v, str):
        return "cmd"
    if not isinstance(v, dict):
        return ""
    if "key" in v or "modifiers" in v or "string" in v:
        return "key"
    if "x" in v or "y" in v or "button" in v:
        return "mouse"
    if "url" in v:
        return "webhook"
    return ""


SequenceElementType = Annotated[
    Union[
        Annotated[KeyPress, Tag("key")],
        Annotated[MouseMove, Tag("mouse")],
        Annotated[str, Tag("cmd")],
        Annotated[WebhookAction, Tag("webhook")],
    ],
    Discriminator(button_action_details_descriminator),
]

SequenceType = list[SequenceElementType]


def button_action_discriminator(v: Any) -> str:
    if v is None:
        return "none"
    if isinstance(v, str):
        if str(v).lower() in [e.value for e in ClientMessageParameterType]:
            return "react"
    if isinstance(v, (list)):
        return "sequence"
    if isinstance(v, (dict)):
        return "single"
    return "none"


class ButtonAction(BaseModel):
    button_id: int
    action: MeetingAction | ActionEnum
    extra: Annotated[
        Union[
            # Teams activation and default actions
            Annotated[None, Tag("none")],
            Annotated[ClientMessageParameterType, Tag("react")],
            Annotated[SequenceType, Tag("sequence")],
            Annotated[SequenceElementType, Tag("single")],
        ],
        Discriminator(button_action_discriminator),
    ] = None


class LedStatus(BaseModel):
    button_id: int
    source: LedStatusSource
    extra: TeamsState | str | None = None
    color_on: LedColor | None = None
    color_off: LedColor | None = None
    read_result: bool = False
    interval: float = 5.0
    timeout: float = 0.5


class VirtualKeypadConfig(BaseModel):
    bind_address: str = "127.0.0.1"
    bind_port: int = 12909


class DeviceInfo(BaseModel):
    vendor_id: int | None = None
    product_id: int | None = None
    serial_number: str | None = None


class LoggingConfig(BaseModel):
    class LogLevel(str, Enum):
        DEBUG = "debug"
        INFO = "info"
        WARNING = "warning"
        ERROR = "error"
        CRITICAL = "critical"

        def to_logging_level(self) -> int:
            return getattr(logging, self.name)

    level: LogLevel = LogLevel.INFO
    submomdules: list[str] = []
    file_enabled: bool = True
    file_path: str | None = None
    file_level: LogLevel = LogLevel.INFO
    file_max_size: int = 3_145_728
    file_backup_count: int = 5
    console_enabled: bool = False
    console_level: LogLevel = LogLevel.INFO


class Config(BaseModel):
    _internal_state: Any = pydantic.PrivateAttr()
    actions: list[ButtonAction]
    longpress_action: list[ButtonAction] = pydantic.Field(
        validation_alias=pydantic.AliasChoices("longpress_action", "double_tap_action"),
    )
    leds: list[LedStatus] = []
    teams_token: str | None = None
    file_path: str | None = None
    virtual_keypad: VirtualKeypadConfig = VirtualKeypadConfig()
    auto_update: bool = True
    device_identifications: list[DeviceInfo] = [
        DeviceInfo(vendor_id=0x1D50, product_id=0x6189, serial_number=None),
        DeviceInfo(vendor_id=7504, product_id=24774, serial_number=None),
        DeviceInfo(vendor_id=4617, product_id=1, serial_number=None),
    ]
    logging: LoggingConfig = LoggingConfig()
    proxy: str | None = None


def create_default_config() -> Config:
    config = Config(
        actions=[
            ButtonAction(button_id=1, action=MeetingAction.ToggleMute),
            ButtonAction(button_id=2, action=MeetingAction.ToggleHand),
            ButtonAction(button_id=3, action=ActionEnum.ACTIVATE_TEAMS),
            ButtonAction(
                button_id=4,
                action=MeetingAction.React,
                extra="like",
            ),
            ButtonAction(button_id=5, action=MeetingAction.LeaveCall),
            ButtonAction(button_id=6, action=MeetingAction.ToggleMute),
            ButtonAction(button_id=7, action=MeetingAction.ToggleHand),
            ButtonAction(button_id=8, action=ActionEnum.ACTIVATE_TEAMS),
            ButtonAction(
                button_id=9,
                action=MeetingAction.React,
                extra="like",
            ),
            ButtonAction(button_id=10, action=MeetingAction.LeaveCall),
        ],
        longpress_action=[
            ButtonAction(button_id=3, action=MeetingAction.ToggleVideo),
            ButtonAction(button_id=8, action=MeetingAction.ToggleVideo),
        ],
        leds=[
            LedStatus(
                button_id=1,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.MUTED,
                color_on=LedColor.RED,
                color_off=LedColor.GREEN,
            ),
            LedStatus(
                button_id=2,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.HAND_RAISED,
                color_on=LedColor.YELLOW,
                color_off=LedColor.BLACK,
            ),
            LedStatus(
                button_id=3,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.VIDEO_ON,
                color_on=LedColor.GREEN,
                color_off=LedColor.RED,
            ),
            LedStatus(
                button_id=5,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.IN_MEETING,
                color_on=LedColor.GREEN,
                color_off=LedColor.BLACK,
            ),
            LedStatus(
                button_id=6,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.MUTED,
                color_on=LedColor.RED,
                color_off=LedColor.GREEN,
            ),
            LedStatus(
                button_id=7,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.HAND_RAISED,
                color_on=LedColor.YELLOW,
                color_off=LedColor.BLACK,
            ),
            LedStatus(
                button_id=8,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.VIDEO_ON,
                color_on=LedColor.RED,
                color_off=LedColor.GREEN,
            ),
            LedStatus(
                button_id=10,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.IN_MEETING,
                color_on=LedColor.GREEN,
                color_off=LedColor.BLACK,
            ),
        ],
        teams_token=None,
        virtual_keypad=VirtualKeypadConfig(
            bind_address="127.0.0.1",
            bind_port=12909,
        ),
    )
    config._internal_state = "default"
    return config


def find_config_file() -> Path:
    file_path = Path(CONFIG_FILENAME)
    home_config_path = (
        Path.home() / os.environ.get("XDG_CONFIG_HOME", ".config") / CONFIG_FILENAME
    )

    if not file_path.exists() and home_config_path.exists():
        file_path = home_config_path

    return file_path


def load_config(file_path: Path | None = None) -> Config:
    if file_path is None:
        file_path = find_config_file()

    try:
        _logger.info("Loading config from file: %s", file_path)
        with open(file_path, "r") as file:
            config_data = yaml.safe_load(file)
        if config_data is None:
            raise yaml.YAMLError("No data in file")
    except FileNotFoundError:
        _logger.info("No config file found, creating default one")
        config = create_default_config()
        config.file_path = str(file_path)
        save_config(config)
        return config

    except (yaml.YAMLError, IOError) as e:
        _logger.info("Error in configuration file: %s", e)
        config = create_default_config()
        config._internal_state = "default_fallback"
        config.file_path = str(file_path)
        return config

    except pydantic.ValidationError as e:
        _logger.info("Configuration errors:")
        for error in e.errors():
            _logger.info(error)

    config_data["file_path"] = str(file_path)
    return Config(**config_data)


def save_config(config: Config, file_path: Path | str | None = None):
    if file_path is None:
        if config.file_path is None:
            config.file_path = str(find_config_file())
        file_path = config.file_path

    config.file_path = file_path  # type: ignore
    if config._internal_state == "default_fallback":
        _logger.error("Not saving default config")
        return
    try:
        with open(file_path, "w") as file:
            yaml.dump(config.model_dump(mode="json"), file)
    except (FileNotFoundError, yaml.YAMLError, IOError):
        _logger.error("Failed to write config to file: %s", file_path)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import asyncio
import logging
import os
import pathlib
import tempfile
import time
from unittest.mock import patch

import pytest
from mutenix.utils import block_parallel
from mutenix.utils import ensure_process_run_once
from mutenix.utils import rate_limited_logger
from mutenix.utils import run_till_some_loop


@pytest.mark.asyncio
async def test_block_parallel_allows_single_execution():
    class TestClass:
        @block_parallel
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()
    result = await test_instance.test_method()
    assert result == "done"


@pytest.mark.asyncio
async def test_block_parallel_blocks_parallel_execution():
    class TestClass:
        @block_parallel
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()

    async def call_method():
        return await test_instance.test_method()

    task1 = asyncio.create_task(call_method())
    task2 = asyncio.create_task(call_method())

    results = await asyncio.gather(task1, task2)
    assert results == ["done", None]


@pytest.mark.asyncio
async def test_block_parallel_resets_after_execution():
    class TestClass:
        @block_parallel
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()
    await test_instance.test_method()
    result = await test_instance.test_method()
    assert result == "done"


@pytest.mark.asyncio
async def test_block_parallel_logs_debug_messages():
    class TestClass:
        @block_parallel
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()

    with patch("mutenix.utils._logger.debug") as mock_debug:
        await test_instance.test_method()
        mock_debug.assert_called_with("block_parallel %s %s", "test_method", False)


@pytest.mark.asyncio
async def test_run_till_some_loop_async():
    class TestClass:
        _run = True

        @run_till_some_loop(sleep_time=0.1)
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()
    result = await test_instance.test_method()
    assert result == "done"


def test_run_till_some_loop_sync():
    class TestClass:
        _run = True
        _first = True

        @run_till_some_loop(sleep_time=0.1)
        def test_method(self):
            time.sleep(0.1)
            if self._first:
                self._first = False
                return None
            return "done"

    test_instance = TestClass()
    result = test_instance.test_method()
    assert result == "done"


@pytest.mark.asyncio
async def test_run_till_some_loop_async_stops_on_condition():
    class TestClass:
        _run = True
        _first = True

        @run_till_some_loop(sleep_time=0.1)
        async def test_method(self):
            await asyncio.sleep(0.1)
            if self._first:
                self._first = False
                return None
            self._run = False
            return "done"

    test_instance = TestClass()
    result = await test_instance.test_method()
    assert result == "done"


def test_run_till_some_loop_sync_stops_on_condition():
    class TestClass:
        _run = True

        @run_till_some_loop(sleep_time=0.1)
        def test_method(self):
            time.sleep(0.1)
            self._run = False
            return "done"

    test_instance = TestClass()
    result = test_instance.test_method()
    assert result == "done"


@pytest.mark.parametrize("limit, interval", [(3, 0.2), (3, 0.4)])
def test_rate_limited_logger(limit, interval):
    logger = logging.getLogger("test_logger")
    logger.setLevel(logging.DEBUG)
    log_handler = logging.StreamHandler()
    log_handler.setLevel(logging.DEBUG)
    logger.addHandler(log_handler)

    @rate_limited_logger(logger, limit=limit, interval=interval)
    def log_message(message):
        logger.debug(message)

    with patch.object(logger, "debug") as mock_debug:
        for _ in range(limit + 2):
            log_message("test message")
        assert mock_debug.call_count == limit

    with patch.object(logger, "warning") as mock_warning:
        time.sleep(interval + 1)
        log_message("test message")
        assert mock_warning.call_count == 1
        assert (
            mock_warning.call_args[0][0]
            == f"Message 'test message' was suppressed {2} times in the last {interval} seconds."
        )


@pytest.fixture
def mock_lock_file():
    lock_file_path = pathlib.Path(tempfile.gettempdir()) / "mutenix.lock"
    if lock_file_path.exists():
        lock_file_path.unlink()
    yield lock_file_path
    if lock_file_path.exists():
        lock_file_path.unlink()


def test_ensure_only_once_first_run(mock_lock_file):
    @ensure_process_run_once()
    def test_func():
        assert mock_lock_file.exists()
        return "Function executed"

    result = test_func()
    assert result == "Function executed"
    assert not mock_lock_file.exists()


def test_ensure_only_once_subsequent_run(mock_lock_file):
    @ensure_process_run_once(lockfile_path=mock_lock_file.parent)
    def test_func():
        return "Function executed"

    with mock_lock_file.open("w") as f:
        f.write(str(os.getpid()))

    with patch("os.kill") as mock_kill:
        with pytest.raises(SystemExit):
            test_func()
        mock_kill.assert_called_once_with(os.getpid(), 0)


@pytest.mark.parametrize("lockfile_exists", [True, False])
def test_ensure_process_run_once(lockfile_exists, mock_lock_file):
    @ensure_process_run_once(lockfile_path=mock_lock_file.parent)
    def test_func():
        return "Function executed"

    if lockfile_exists:
        with mock_lock_file.open("w") as f:
            f.write(str(os.getpid()))

    if lockfile_exists:
        with patch("psutil.pid_exists", return_value=True):
            with pytest.raises(SystemExit):
                test_func()
    else:
        result = test_func()
        assert result == "Function executed"
        assert not mock_lock_file.exists()


def test_ensure_process_run_once_stale_lockfile(mock_lock_file):
    @ensure_process_run_once(lockfile_path=mock_lock_file.parent)
    def test_func():
        return "Function executed"

    with mock_lock_file.open("w") as f:
        f.write("invalid_pid")

    result = test_func()
    assert result == "Function executed"
    assert not mock_lock_file.exists()


def test_ensure_process_run_once_lockfile_cleanup(mock_lock_file):
    @ensure_process_run_once(lockfile_path=mock_lock_file.parent)
    def test_func():
        return "Function executed"

    result = test_func()
    assert result == "Function executed"
    assert not mock_lock_file.exists()



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import asyncio
import functools
import logging
import pathlib
import platform
import tempfile
import time


_logger = logging.getLogger(__name__)

if platform.system().lower() == "windows":  # pragma: no cover
    from mutenix.utils.windows import bring_teams_to_foreground, ensure_process_run_once
elif platform.system().lower() == "linux":  # pragma: no cover
    from mutenix.utils.linux import bring_teams_to_foreground, ensure_process_run_once
elif platform.system().lower() == "darwin":  # pragma: no cover
    from mutenix.utils.darwin import bring_teams_to_foreground, ensure_process_run_once
else:
    _logger.error("Platform not supported")

    def bring_teams_to_foreground() -> None:
        pass

    def ensure_process_run_once(
        lockfile_path: pathlib.Path = pathlib.Path(tempfile.gettempdir())
        / "mutenix.lock",
    ):
        def wrapper(func):
            return func

        return wrapper


def run_loop(func):
    if asyncio.iscoroutinefunction(func):

        async def wrapper(self, *args, **kwargs):
            while self._run:
                await func(self, *args, **kwargs)
                await asyncio.sleep(0)

    else:
        raise Exception("only for async functions")  # pragma: no cover
    return wrapper


def block_parallel(func):
    """Blocks parallel calls to the function."""
    func._already_running = False

    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        _logger.debug("block_parallel %s %s", func.__name__, func._already_running)
        if func._already_running:
            while func._already_running:
                await asyncio.sleep(0.1)
            return
        func._already_running = True
        result = await func(self, *args, **kwargs)
        func._already_running = False
        return result

    return wrapper


def run_till_some_loop(sleep_time: float = 0):
    def decorator(func):
        if asyncio.iscoroutinefunction(func):

            async def wrapper(self, *args, **kwargs):
                while self._run:
                    some = await func(self, *args, **kwargs)
                    if some:
                        return some
                    if sleep_time > 0:
                        await asyncio.sleep(sleep_time)
        else:

            def wrapper(self, *args, **kwargs):
                while self._run:
                    some = func(self, *args, **kwargs)
                    if some:
                        return some
                    if sleep_time > 0:
                        time.sleep(sleep_time)

        return wrapper

    return decorator


def rate_limited_logger(logger, limit=3, interval=10):
    """
    A decorator to limit repeated log messages.

    Args:
        logger (logging.Logger): The logger instance to use.
        limit (int): The number of allowed repeated log messages.
        interval (int): The time interval in seconds within which repeated log messages are limited.

    Returns:
        function: The wrapped logging function.
    """

    def decorator(func):
        last_logged = {}
        log_count = {}

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            message = args[0] if args else ""
            current_time = time.monotonic()

            if message not in last_logged:
                last_logged[message] = 0
                log_count[message] = 0

            if current_time - last_logged[message] > interval:
                if log_count[message] > limit:
                    logger.warning(
                        f"Message '{message}' was suppressed {log_count[message] - limit} times in the last {interval} seconds.",
                    )
                log_count[message] = 0

            if log_count[message] < limit:
                func(*args, **kwargs)
                last_logged[message] = current_time
                log_count[message] += 1
            else:
                log_count[message] += 1
                last_logged[message] = current_time

        return wrapper

    return decorator



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import argparse
from pathlib import Path
from unittest.mock import patch

import pytest
from mutenix.__main__ import list_devices
from mutenix.__main__ import main
from mutenix.__main__ import parse_arguments


def test_parse_arguments_no_update_file():
    test_args = []
    with patch("sys.argv", ["__main__.py"] + test_args):
        args = parse_arguments()
        assert args.update_file is None


def test_parse_arguments_with_update_file():
    test_args = ["--update-file", "path/to/update.tar.gz"]
    with patch("sys.argv", ["__main__.py"] + test_args):
        args = parse_arguments()
        assert args.update_file == "path/to/update.tar.gz"


def test_parse_arguments_list_devices():
    test_args = ["--list-devices"]
    with patch("sys.argv", ["__main__.py"] + test_args):
        args = parse_arguments()
        assert args.list_devices is True


def test_parse_arguments_config():
    test_args = ["--config", "path/to/config.yaml"]
    with patch("sys.argv", ["__main__.py"] + test_args):
        args = parse_arguments()
        assert args.config == Path("path/to/config.yaml")


@pytest.fixture
def mock_macropad():
    with patch("mutenix.__main__.Macropad", autospec=True) as MockMacropad:
        yield MockMacropad.return_value


@pytest.fixture
def mock_check_for_self_update():
    with patch("mutenix.__main__.check_for_self_update") as mock_update:
        yield mock_update


@pytest.fixture
def mock_run_trayicon():
    with patch("mutenix.__main__.run_trayicon") as mock_trayicon:
        yield mock_trayicon


@pytest.fixture
def mock_signal():
    with patch("signal.signal") as mock_signal:
        yield mock_signal


def test_main_no_update_file(
    mock_macropad,
    mock_check_for_self_update,
    mock_run_trayicon,
    mock_signal,
):
    args = argparse.Namespace(update_file=None, list_devices=False, config=None)

    with (
        patch("threading.Thread.start", autospec=True) as mock_thread_start,
        patch("threading.Thread.join", autospec=True) as mock_thread_join,
        patch("asyncio.run", autospec=True),
    ):
        main(args)

        mock_check_for_self_update.assert_called_once()
        mock_signal.assert_called_once()
        mock_thread_start.assert_called_once()
        mock_run_trayicon.assert_called_once_with(mock_macropad)
        mock_thread_join.assert_called_once()


def test_main_with_update_file(mock_macropad, mock_check_for_self_update, mock_signal):
    args = argparse.Namespace(
        update_file="path/to/update.tar.gz",
        list_devices=False,
        config=None,
    )

    with patch("asyncio.run", autospec=True) as mock_asyncio_run:
        main(args)

        mock_check_for_self_update.assert_called_once()
        mock_signal.assert_called_once()
        mock_asyncio_run.assert_called_once()


def test_main_list_devices(mock_signal):
    args = argparse.Namespace(update_file=None, list_devices=True, config=None)

    with patch("hid.enumerate", autospec=True) as mock_hid_enumerate:
        main(args)

        mock_signal.assert_not_called()
        mock_hid_enumerate.assert_called_once()


def test_list_devices(capsys):
    with patch("hid.enumerate", autospec=True) as mock_hid:
        # Mock the return value of hid.enumerate
        mock_hid.return_value = [
            {
                "path": "device1",
                "vendor_id": 1234,
                "product_id": 5678,
                "product_string": "mutenix",
            },
            {
                "path": "device2",
                "vendor_id": 8765,
                "product_id": 4321,
                "product_string": "selse",
            },
        ]

        # Call the function
        list_devices()

        # Capture the output
        captured = capsys.readouterr()

        # Check if the output contains the mocked devices
        assert "device1" in captured.out
        assert "device2" in captured.out
        assert "1234" in captured.out
        assert "5678" in captured.out
        assert "8765" in captured.out
        assert "4321" in captured.out



HEADER_SIZE = 8
MAX_CHUNK_SIZE = 60 - HEADER_SIZE
DATA_TRANSFER_SLEEP_TIME = 1
STATE_CHANGE_SLEEP_TIME = 0.5
WAIT_FOR_REQUESTS_SLEEP_TIME = STATE_CHANGE_SLEEP_TIME
HID_REPORT_ID_COMMUNICATION = 1
HID_REPORT_ID_TRANSFER = 2

HID_COMMAND_PREPARE_UPDATE = 0xE0
HID_COMMAND_RESET = 0xE1



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import asyncio
import shlex
import time
from collections import defaultdict
from unittest.mock import ANY
from unittest.mock import AsyncMock
from unittest.mock import Mock
from unittest.mock import mock_open
from unittest.mock import patch

import pytest
from mutenix.config import create_default_config
from mutenix.config import LedStatusSource
from mutenix.config import WebhookAction
from mutenix.hid_commands import LedColor
from mutenix.hid_commands import SetLed
from mutenix.hid_commands import Status
from mutenix.hid_commands import VersionInfo
from mutenix.macropad import Macropad
from mutenix.teams_messages import ClientMessageParameter
from mutenix.teams_messages import ClientMessageParameterType
from mutenix.teams_messages import MeetingAction
from mutenix.teams_messages import MeetingPermissions
from mutenix.teams_messages import MeetingState
from mutenix.teams_messages import MeetingUpdate
from mutenix.teams_messages import ServerMessage

try:
    from pynput.keyboard import Key
    from pynput.mouse import Button
except ImportError:
    Key = Mock()
    Button = Mock()


@pytest.fixture
def macropad():
    with (
        patch("mutenix.macropad.HidDevice") as MockHidDevice,
        patch("mutenix.macropad.WebSocketClient") as MockWebSocketClient,
        patch("mutenix.macropad.VirtualMacropad") as MockVirtualMacropad,
        patch("mutenix.config.save_config"),
    ):
        MockHidDevice.return_value = Mock()
        MockWebSocketClient.return_value = Mock()
        MockVirtualMacropad.return_value = Mock()
        return Macropad(create_default_config())


@pytest.mark.asyncio
async def test_hid_callback_status(macropad):
    msg = Status(bytes([1, 1, 0, 0, 1]))
    macropad._websocket.send_message = AsyncMock()
    await macropad._hid_callback(msg)
    macropad._websocket.send_message.assert_called_once()
    assert (
        macropad._websocket.send_message.call_args[0][0].action
        == MeetingAction.ToggleMute
    )


@pytest.mark.asyncio
async def test_hid_callback_version_info(macropad):
    msg = VersionInfo(bytes([1, 0, 0, 2]))
    macropad._version_seen = None
    with patch(
        "mutenix.macropad.check_for_device_update",
    ) as mock_check_for_device_update:
        await macropad._hid_callback(msg)
        mock_check_for_device_update.assert_called_once_with(ANY, msg, ANY)


@pytest.mark.asyncio
async def test_hid_callback_version_info_only_once(macropad):
    msg = VersionInfo(bytes([1, 0, 0, 2]))
    macropad._version_seen = None
    await macropad._hid_callback(msg)
    with patch(
        "mutenix.macropad.check_for_device_update",
    ) as mock_check_for_device_update:
        await macropad._hid_callback(msg)
        mock_check_for_device_update.assert_not_called()


@pytest.mark.asyncio
async def test_teams_callback_token_refresh(macropad):
    msg = ServerMessage(tokenRefresh="new_token")
    macropad._current_state = None
    with patch("builtins.open", mock_open()) as mock_file:
        await macropad._teams_callback(msg)
        mock_file.assert_called_once()
        assert str(mock_file.call_args[0][0]).endswith("mutenix.yaml")
        mock_file().write.assert_any_call("new_token")


@pytest.mark.asyncio
async def test_teams_callback_token_refresh_save_failed(macropad):
    msg = ServerMessage(tokenRefresh="new_token")
    macropad._current_state = None
    with (
        patch("builtins.open", mock_open()) as mock_file,
        patch("mutenix.macropad._logger.error"),
    ):
        mock_file().write.side_effect = IOError
        await macropad._teams_callback(msg)
        mock_file.assert_called()
        assert str(mock_file.call_args[0][0]).endswith("mutenix.yaml")


@pytest.mark.asyncio
async def test_update_device_status(macropad):
    macropad._current_state = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=True,
                isMuted=True,
                isHandRaised=False,
                isVideoOn=True,
            ),
            meetingPermissions=MeetingPermissions(canLeave=True),
        ),
    )

    def send_msg(msg):
        future = asyncio.get_event_loop().create_future()
        future.set_result(None)
        assert isinstance(msg, SetLed)
        return future

    macropad._device.send_msg = Mock(side_effect=send_msg)
    macropad._virtual_macropad.send_msg = Mock()
    await macropad._update_device_status()
    assert macropad._device.send_msg.call_count == 8
    assert macropad._virtual_macropad.send_msg.call_count == 8


@pytest.mark.asyncio
async def test_update_device_status_not_in_meeting(macropad):
    macropad._current_state = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=False,
                isMuted=False,
                isHandRaised=False,
                isVideoOn=False,
            ),
            meetingPermissions=MeetingPermissions(canLeave=False),
        ),
    )

    def send_msg(msg):
        future = asyncio.get_event_loop().create_future()
        future.set_result(None)
        assert isinstance(msg, SetLed)
        return future

    macropad._device.send_msg = Mock(side_effect=send_msg)
    macropad._virtual_macropad.send_msg = AsyncMock()
    await macropad._update_device_status()
    assert macropad._device.send_msg.call_count == 8
    assert macropad._virtual_macropad.send_msg.call_count == 8


@pytest.mark.parametrize(
    "msg_bytes, expected_action, should_call",
    [
        (bytes([1, 1, 0, 0, 1]), MeetingAction.ToggleMute, True),
        (bytes([2, 1, 0, 0, 1]), MeetingAction.ToggleHand, True),
        (bytes([4, 1, 0, 0, 1]), MeetingAction.React, True),
        (bytes([5, 1, 0, 0, 1]), MeetingAction.LeaveCall, True),
        (bytes([1, 0, 0, 0, 1]), None, False),
        (bytes([2, 0, 0, 0, 1]), None, False),
        (bytes([4, 0, 0, 0, 1]), None, False),
        (bytes([5, 0, 0, 0, 1]), None, False),
        (bytes([1, 1, 0, 0, 0]), None, False),
        (bytes([2, 1, 0, 0, 0]), None, False),
        (bytes([4, 1, 0, 0, 0]), None, False),
        (bytes([5, 1, 0, 0, 0]), None, False),
    ],
)
@pytest.mark.asyncio
async def test_hid_callback_parametrized(
    macropad,
    msg_bytes,
    expected_action: MeetingAction,
    should_call,
):
    msg = Status(msg_bytes)
    macropad._websocket.send_message = AsyncMock()

    await macropad._hid_callback(msg)
    if should_call:
        macropad._websocket.send_message.assert_called_once()
        if expected_action:
            assert (
                macropad._websocket.send_message.call_args[0][0].action.name
                == expected_action.name
            )
    else:
        macropad._websocket.send_message.assert_not_called()


@pytest.mark.asyncio
async def test_hid_callback_invalid_button(macropad):
    msg = Status([11, 1, 0, 0, 1])

    await macropad._hid_callback(msg)
    macropad._websocket.send_message.assert_not_called()


class IdentifierWithoutToken:
    def __eq__(self, value):
        if value.token:
            return False
        return True


@pytest.mark.asyncio
async def test_setup_without_existing_token():
    with patch("builtins.open", mock_open(read_data="")) as mock_file:
        mock_file.side_effect = FileNotFoundError
        with patch("mutenix.macropad.HidDevice"):
            with patch("mutenix.macropad.WebSocketClient") as MockWebSocketClient:
                with patch("mutenix.macropad.VirtualMacropad"):
                    macropad = Macropad(create_default_config())
                    macropad._setup()
                    MockWebSocketClient.assert_called_with(
                        ANY,
                        IdentifierWithoutToken(),
                    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "status, expected_action, expected_parameters",
    [
        (Status(bytes([1, 1, 0, 0, 1])), MeetingAction.ToggleMute, None),
        (Status(bytes([2, 1, 0, 0, 1])), MeetingAction.ToggleHand, None),
        # bring_teams_to_foreground
        (Status(bytes([3, 1, 0, 0, 1])), None, None),
        (
            Status(bytes([3, 1, 1, 0, 1])),
            MeetingAction.ToggleVideo,
            None,
        ),  # longpressed
        (
            Status(bytes([4, 1, 0, 0, 1])),
            MeetingAction.React,
            ClientMessageParameter(type_=ClientMessageParameterType.ReactLike),
        ),
        (Status(bytes([5, 1, 0, 0, 1])), MeetingAction.LeaveCall, None),
        (Status(bytes([1, 0, 0, 0, 1])), None, None),
    ],
)
async def test_send_status(macropad, status, expected_action, expected_parameters):
    macropad._websocket.send_message = AsyncMock()
    with patch(
        "mutenix.macropad.bring_teams_to_foreground",
    ) as mock_bring_teams_to_foreground:
        await macropad._send_status(status)
        if expected_action:
            macropad._websocket.send_message.assert_called_once()
            client_message = macropad._websocket.send_message.call_args[0][0]
            assert client_message.action == expected_action
            if expected_parameters:
                assert client_message.parameters == expected_parameters
        else:
            macropad._websocket.send_message.assert_not_called()
            if status.button == 3 and status.triggered and status.released:
                mock_bring_teams_to_foreground.assert_called_once()
            else:
                mock_bring_teams_to_foreground.assert_not_called()


@pytest.mark.asyncio
async def test_process(macropad):
    macropad._device.process = AsyncMock()
    macropad._websocket.process = AsyncMock()
    macropad._virtual_macropad.process = AsyncMock()

    async def stop_macropad():
        await asyncio.sleep(0.2)
        await macropad.stop()

    asyncio.create_task(stop_macropad())

    await macropad.process()

    macropad._device.process.assert_called_once()
    macropad._websocket.process.assert_called_once()
    macropad._virtual_macropad.process.assert_called_once()


@pytest.mark.asyncio
async def test_process_with_exception(macropad):
    macropad._device.process = AsyncMock(side_effect=Exception("Device error"))
    macropad._websocket.process = AsyncMock(return_value=None)
    macropad._virtual_macropad.process = AsyncMock(return_value=None)

    macropad._check_status = AsyncMock()

    with patch("mutenix.macropad._logger.error") as mock_logger_error:
        await macropad.process()
        mock_logger_error.assert_called_with("Error in Macropad process: %s", ANY)

    macropad._device.process.assert_called_once()
    macropad._websocket.process.assert_called_once()
    macropad._virtual_macropad.process.assert_called_once()


@pytest.mark.asyncio
async def test_manual_update_success(macropad):
    update_file = "update.bin"
    macropad._device.wait_for_device = AsyncMock()
    with patch("builtins.open", mock_open(read_data=b"update_data")) as mock_file:
        with patch(
            "mutenix.macropad.perform_upgrade_with_file",
        ) as mock_perform_upgrade:
            await macropad.manual_update(update_file)
            macropad._device.wait_for_device.assert_called_once()
            mock_file.assert_called_once_with(update_file, "rb")
            mock_perform_upgrade.assert_called_once_with(
                macropad._device.raw,
                mock_file(),
            )


@pytest.mark.asyncio
async def test_manual_update_file_not_found(macropad):
    update_file = "non_existent_update.bin"
    macropad._device.wait_for_device = AsyncMock()
    with patch("builtins.open", mock_open()) as mock_file:
        mock_file.side_effect = FileNotFoundError
        with patch(
            "mutenix.macropad.perform_upgrade_with_file",
        ) as mock_perform_upgrade:
            with pytest.raises(FileNotFoundError):
                await macropad.manual_update(update_file)
            macropad._device.wait_for_device.assert_called_once()
            mock_file.assert_called_once_with(update_file, "rb")
            mock_perform_upgrade.assert_not_called()


@pytest.mark.asyncio
async def test_manual_update_io_error(macropad):
    update_file = "update.bin"
    macropad._device.wait_for_device = AsyncMock()
    with patch("builtins.open", mock_open()) as mock_file:
        mock_file.side_effect = IOError
        with patch(
            "mutenix.macropad.perform_upgrade_with_file",
        ) as mock_perform_upgrade:
            with pytest.raises(IOError):
                await macropad.manual_update(update_file)
            macropad._device.wait_for_device.assert_called_once()
            mock_file.assert_called_once_with(update_file, "rb")
            mock_perform_upgrade.assert_not_called()


@pytest.mark.asyncio
async def test_stop():
    macropad = Macropad(create_default_config())
    macropad._device = AsyncMock()
    macropad._websocket = AsyncMock()
    macropad._virtual_macropad = AsyncMock()

    await macropad.stop()

    macropad._device.stop.assert_called_once()
    macropad._websocket.stop.assert_called_once()
    macropad._virtual_macropad.stop.assert_called_once()


@pytest.mark.asyncio
async def test_update_device_status_teams_source_in_meeting(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.TEAMS,
            button_id=1,
            extra="is-muted",
            color_on="red",
            color_off="green",
        ),
    ]
    macropad._current_state = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=True,
                isMuted=True,
                isHandRaised=False,
                isVideoOn=True,
            ),
            meetingPermissions=MeetingPermissions(canLeave=True),
        ),
    )

    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.RED))
    macropad._virtual_macropad.send_msg.assert_called_once_with(SetLed(1, LedColor.RED))


@pytest.mark.asyncio
async def test_update_device_status_teams_source_not_in_meeting(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.TEAMS,
            button_id=1,
            extra="is-muted",
            color_on="red",
            color_off="green",
        ),
    ]
    macropad._current_state = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=False,
                isMuted=False,
                isHandRaised=False,
                isVideoOn=False,
            ),
            meetingPermissions=MeetingPermissions(canLeave=False),
        ),
    )

    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.BLACK))
    macropad._virtual_macropad.send_msg.assert_called_once_with(
        SetLed(1, LedColor.BLACK),
    )


@pytest.mark.asyncio
async def test_update_device_status_cmd_source_with_result(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.CMD,
            button_id=1,
            extra="echo blue",
            interval=0,
            read_result=True,
            timeout=1.0,
        ),
    ]
    macropad._last_status_check = defaultdict(int)
    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    with patch("asyncio.to_thread", return_value="blue"):
        await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.BLUE))
    macropad._virtual_macropad.send_msg.assert_called_once_with(
        SetLed(1, LedColor.BLUE),
    )


@pytest.mark.asyncio
async def test_update_device_status_cmd_source_without_result(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.CMD,
            button_id=1,
            extra="exit 0",
            interval=0,
            read_result=False,
            color_on="yellow",
            color_off="black",
            timeout=1.0,
        ),
    ]
    macropad._last_status_check = defaultdict(int)
    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    with patch("asyncio.to_thread", return_value=0):
        await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.YELLOW))
    macropad._virtual_macropad.send_msg.assert_called_once_with(
        SetLed(1, LedColor.YELLOW),
    )


@pytest.mark.asyncio
async def test_update_device_status_cmd_source_interval_not_elapsed(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.CMD,
            button_id=1,
            extra="exit 0",
            interval=10,
            read_result=False,
            color_on="yellow",
            color_off="black",
        ),
    ]
    macropad._last_status_check = defaultdict(lambda: time.time())
    macropad._device.send_msg = AsyncMock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    await macropad._update_device_status()

    macropad._device.send_msg.assert_not_called()
    macropad._virtual_macropad.send_msg.assert_not_called()


@pytest.mark.asyncio
async def test_do_check_status_update_device_status_called(macropad):
    macropad._update_device_status = AsyncMock()
    macropad._checktime = time.time() - 11  # Ensure the time condition is met

    await macropad._do_check_status()

    macropad._update_device_status.assert_called_once()


@pytest.mark.asyncio
async def test_do_check_status_sleep_called(macropad):
    macropad._update_device_status = AsyncMock()
    macropad._checktime = time.time() - 11  # Ensure the time condition is met

    with patch("asyncio.sleep", return_value=None) as mock_sleep:
        await macropad._do_check_status()
        mock_sleep.assert_called_once_with(0.1)


def test_activate_serial_console(macropad):
    macropad._device.send_msg = Mock()
    macropad.activate_serial_console()
    macropad._device.send_msg.assert_called_once()
    sent_message = macropad._device.send_msg.call_args[0][0]
    assert sent_message._activate_debug == 2
    assert sent_message._activate_filesystem == 0


def test_deactivate_serial_console(macropad):
    macropad._device.send_msg = Mock()
    macropad.deactivate_serial_console()
    macropad._device.send_msg.assert_called_once()
    sent_message = macropad._device.send_msg.call_args[0][0]
    assert sent_message._activate_debug == 1
    assert sent_message._activate_filesystem == 0


def test_activate_filesystem(macropad):
    macropad._device.send_msg = Mock()
    macropad.activate_filesystem()
    macropad._device.send_msg.assert_called_once()
    sent_message = macropad._device.send_msg.call_args[0][0]
    assert sent_message._activate_filesystem == 2


@pytest.mark.parametrize(
    "extra, expected_calls",
    [
        ({"key": "a"}, [("tap", "a")]),
        ({"key": "enter"}, [("tap", Key.enter)]),
        (
            {"modifiers": ["ctrl", "shift"], "key": "a"},
            [("pressed", Key.ctrl), ("pressed", Key.shift), ("tap", "a")],
        ),
        ({"string": "hello"}, [("type", "hello")]),
    ],
)
@pytest.mark.skipif(isinstance(Key, Mock), reason="pynput not supported on plattform")
def test_keypress(extra, expected_calls):
    with patch("mutenix.macropad.Controller") as MockController:
        mock_keyboard = MockController.return_value
        macropad = Macropad(create_default_config())
        macropad._keypress(extra)

        if "key" in extra:
            if len(extra["key"]) == 1:
                mock_keyboard.tap.assert_any_call(extra["key"])
            else:
                mock_keyboard.tap.assert_any_call(getattr(Key, extra["key"]))
        elif "string" in extra:
            mock_keyboard.type.assert_called_once_with(extra["string"])

        for call in expected_calls:
            if call[0] == "pressed":
                mock_keyboard.pressed.assert_any_call(call[1])
            elif call[0] == "tap":
                mock_keyboard.tap.assert_any_call(call[1])
            elif call[0] == "type":
                mock_keyboard.type.assert_any_call(call[1])


@pytest.mark.skipif(isinstance(Key, Mock), reason="pynput not supported on plattform")
def test_keypress_pynput_not_supported():
    with patch("mutenix.macropad.Controller", None):
        macropad = Macropad(create_default_config())
        with patch("mutenix.macropad._logger.error") as mock_logger_error:
            macropad._keypress({"key": "a"})
            mock_logger_error.assert_called_once_with(
                "pynput not supported, cannot send keypress",
            )


@pytest.mark.skipif(isinstance(Key, Mock), reason="pynput not supported on plattform")
def test_keypress_invalid_key():
    with patch("mutenix.macropad.Controller") as MockController:
        mock_keyboard = MockController.return_value
        macropad = Macropad(create_default_config())
        macropad._keypress({"key": "invalid_key"})
        mock_keyboard.press.assert_not_called()
        mock_keyboard.release.assert_not_called()


@pytest.mark.parametrize(
    "extra, expected_calls",
    [
        ({"action": "move", "x": 10, "y": 20}, [("move", 10, 20)]),
        ({"action": "set", "x": 30, "y": 40}, [("position", (30, 40))]),
        (
            {"action": "click", "button": "left", "count": 2},
            [("click", Button.left, 2)],
        ),
        ({"action": "press", "button": "right"}, [("press", Button.right)]),
        ({"action": "release", "button": "middle"}, [("release", Button.middle)]),
    ],
)
@pytest.mark.skipif(
    isinstance(Button, Mock),
    reason="pynput not supported on plattform",
)
def test_mousemove(extra, expected_calls):
    with patch("mutenix.macropad.MouseController") as MockMouseController:
        mock_mouse = MockMouseController.return_value
        macropad = Macropad(create_default_config())
        macropad._mousemove(extra)

        for call in expected_calls:
            if call[0] == "move":
                mock_mouse.move.assert_called_once_with(call[1], call[2])
            elif call[0] == "position":
                assert mock_mouse.position == call[1]
            elif call[0] == "click":
                mock_mouse.click.assert_called_once_with(call[1], call[2])
            elif call[0] == "press":
                mock_mouse.press.assert_called_once_with(call[1])
            elif call[0] == "release":
                mock_mouse.release.assert_called_once_with(call[1])


@pytest.mark.skipif(
    isinstance(Button, Mock),
    reason="pynput not supported on plattform",
)
def test_mousemove_pynput_not_supported():
    with patch("mutenix.macropad.MouseController", None):
        macropad = Macropad(create_default_config())
        with patch("mutenix.macropad._logger.error") as mock_logger_error:
            macropad._mousemove({"action": "move", "x": 10, "y": 20})
            mock_logger_error.assert_called_once_with(
                "pynput not supported, cannot send mousemove",
            )


@pytest.mark.skipif(
    isinstance(Button, Mock),
    reason="pynput not supported on plattform",
)
def test_mousemove_invalid_action():
    with patch("mutenix.macropad.MouseController") as MockMouseController:
        mock_mouse = MockMouseController.return_value
        macropad = Macropad(create_default_config())
        macropad._mousemove({"action": "invalid_action"})
        mock_mouse.move.assert_not_called()
        mock_mouse.position = (0, 0)
        mock_mouse.click.assert_not_called()
        mock_mouse.press.assert_not_called()
        mock_mouse.release.assert_not_called()


@pytest.mark.parametrize(
    "color, expected_led_color",
    [
        ("red", LedColor.RED),
        ("green", LedColor.GREEN),
        ("blue", LedColor.BLUE),
        ("yellow", LedColor.YELLOW),
        ("invalid_color", LedColor.GREEN),  # Default to GREEN for invalid colors
    ],
)
def test_map_led_color(color, expected_led_color):
    macropad = Macropad(create_default_config())
    result = macropad._map_led_color(color)
    assert result == expected_led_color


@pytest.mark.asyncio
async def test_reload_config(macropad):
    with (
        patch("mutenix.macropad.load_config") as mock_load_config,
        patch("mutenix.macropad._logger.info") as mock_logger_info,
        patch.object(macropad, "_setup_buttons") as mock_setup_buttons,
        patch.object(macropad, "_update_device_status") as mock_update_device_status,
    ):
        mock_load_config.return_value = create_default_config()
        macropad.reload_config()
        mock_load_config.assert_called_once()
        mock_setup_buttons.assert_called_once()
        mock_update_device_status.assert_called_once_with(force=True)
        mock_logger_info.assert_any_call("Reloading config")
        mock_logger_info.assert_any_call("Config reloaded")


@pytest.mark.parametrize(
    "extra, expected_method, expected_url, expected_json, expected_headers",
    [
        (
            {
                "method": "POST",
                "url": "http://example.com",
                "data": {"key": "value"},
                "headers": {"Content-Type": "application/json"},
            },
            "POST",
            "http://example.com",
            {"key": "value"},
            {"Content-Type": "application/json"},
        ),
        (
            {"url": "http://example.com"},
            "GET",
            "http://example.com",
            None,
            {},
        ),
        (
            {
                "method": "PUT",
                "url": "http://example.com",
                "headers": {"Authorization": "Bearer token"},
            },
            "PUT",
            "http://example.com",
            None,
            {"Authorization": "Bearer token"},
        ),
    ],
)
def test_perform_webhook(
    extra,
    expected_method,
    expected_url,
    expected_json,
    expected_headers,
    macropad,
):
    with patch("mutenix.macropad.requests.request") as mock_request:
        extra = WebhookAction(**extra)
        macropad._perform_webhook(extra)
        mock_request.assert_called_once_with(
            expected_method,
            expected_url,
            json=expected_json,
            headers=expected_headers,
        )


@pytest.mark.asyncio
async def test_teams_callback_meeting_update(macropad):
    msg = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=True,
                isMuted=False,
                isHandRaised=False,
                isVideoOn=True,
            ),
            meetingPermissions=MeetingPermissions(canLeave=True),
        ),
    )
    macropad._current_state = None
    await macropad._teams_callback(msg)
    assert macropad._current_state == msg


@pytest.mark.asyncio
async def test_teams_callback_update_device_status_called(macropad):
    msg = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=True,
                isMuted=False,
                isHandRaised=False,
                isVideoOn=True,
            ),
            meetingPermissions=MeetingPermissions(canLeave=True),
        ),
    )
    macropad._current_state = None
    with patch.object(macropad, "_update_device_status") as mock_update_device_status:
        await macropad._teams_callback(msg)
        mock_update_device_status.assert_called_once()


@pytest.mark.asyncio
async def test_update_device_status_webhook_source_with_color(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.WEBHOOK,
            button_id=1,
            extra=None,
            color_on="red",
            color_off="green",
        ),
    ]
    macropad._virtual_macropad.get_led_status = Mock(return_value="red")
    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.RED))
    macropad._virtual_macropad.send_msg.assert_called_once_with(SetLed(1, LedColor.RED))


@pytest.mark.asyncio
async def test_update_device_status_webhook_source_with_invalid_color(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.WEBHOOK,
            button_id=1,
            extra=None,
            color_on="red",
            color_off="green",
        ),
    ]
    macropad._virtual_macropad.get_led_status = Mock(return_value="invalid_color")
    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.GREEN))
    macropad._virtual_macropad.send_msg.assert_called_once_with(
        SetLed(1, LedColor.GREEN),
    )


@pytest.mark.parametrize(
    "command, expected_stdout, expected_stderr, expected_returncode",
    [
        ("echo Hello", "Hello\n", "", 0),
        ("exit 1", "", "", 1),
    ],
)
def test_do_run_command(
    command,
    expected_stdout,
    expected_stderr,
    expected_returncode,
    macropad,
):
    with patch("mutenix.macropad.subprocess.run") as mock_run:
        mock_run.return_value.stdout = expected_stdout
        mock_run.return_value.stderr = expected_stderr
        mock_run.return_value.returncode = expected_returncode
        with patch("mutenix.macropad._logger.debug") as mock_logger_debug:
            macropad._do_run_command(command)

            mock_run.assert_called_once_with(
                shlex.split(command),
                capture_output=True,
                text=True,
            )
            mock_logger_debug.assert_any_call("Running command: %s", command)
            mock_logger_debug.assert_any_call("Command output: %s", expected_stdout)
            mock_logger_debug.assert_any_call("Command error: %s", expected_stderr)
            mock_logger_debug.assert_any_call(
                "Command return code: %s",
                expected_returncode,
            )



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import asyncio
import logging
import webbrowser
from pathlib import Path
from typing import Any

from mutenix.macropad import Macropad
from mutenix.version import MAJOR
from mutenix.version import MINOR
from mutenix.version import PATCH
from PIL import Image

my_icon: Any | None = None
_logger = logging.getLogger(__name__)


def load_image(file_name):
    try:
        file_path = Path(__file__).parent / "assets" / file_name
        return Image.open(file_path)
    except Exception as e:
        _logger.error("Error loading image: %s", e)
        return Image.new("RGB", (64, 64), "red")


def run_trayicon(macropad: Macropad):  # pragma: no cover
    global my_icon
    from pystray import Icon as icon, Menu as menu, MenuItem as item

    def open_url(endpoint=""):
        def open_url(icon, item):
            address = (
                macropad.virtual_keypad_address
                if macropad.virtual_keypad_address != "0.0.0.0"
                else "127.0.0.1"
            )
            webbrowser.open(
                f"http://{address}:{macropad.virtual_keypad_port}{endpoint}",
            )

        return open_url

    def activate_serial_console(icon, item):
        macropad.activate_serial_console()

    def deactivate_serial_console(icon, item):
        macropad.deactivate_serial_console()

    def activate_filesystem(icon, item):
        macropad.activate_filesystem()

    def quit_macropad(icon, item):
        icon.stop()
        asyncio.run(macropad.stop())

    def nothing(icon, item):
        pass

    my_icon = icon(
        "MUTENIX",
        load_image("icon_all_red_64.png"),
        menu=menu(
            item(
                "Open Virtual Macropad",
                open_url("/"),
            ),
            item(
                "Reload Config",
                macropad.reload_config,
            ),
            item(
                "Teams connected",
                nothing,
                checked=lambda x: macropad.teams_connected,
                enabled=False,
            ),
            item(
                "Device connected",
                nothing,
                checked=lambda x: macropad.device_connected,
                enabled=False,
            ),
            item(
                "Show Config",
                open_url("/config"),
            ),
            item(
                "Help",
                open_url("/help"),
            ),
            item(
                "About",
                open_url("/about"),
            ),
            item(
                "Device",
                open_url("/device"),
            ),
            item(
                "Debug Options",
                menu(
                    item("Activate Serial Console", activate_serial_console),
                    item("Deactivate Serial Console", deactivate_serial_console),
                    item("Activate Filesystem", activate_filesystem),
                ),
            ),
            item(
                "Quit",
                quit_macropad,
            ),
            item(
                f"Version {MAJOR}.{MINOR}.{PATCH}",
                nothing,
                enabled=False,
            ),
        ),
    )
    my_icon.run()



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
from enum import Enum
from typing import ClassVar
from typing import Optional

from pydantic import BaseModel
from pydantic import Field


class MeetingPermissions(BaseModel):
    """Permissions for the meeting communicated by teams."""

    can_toggle_mute: bool = Field(False, alias="canToggleMute")
    can_toggle_video: bool = Field(False, alias="canToggleVideo")
    can_toggle_hand: bool = Field(False, alias="canToggleHand")
    can_toggle_blur: bool = Field(False, alias="canToggleBlur")
    can_leave: bool = Field(False, alias="canLeave")
    can_react: bool = Field(False, alias="canReact")
    can_toggle_share_tray: bool = Field(False, alias="canToggleShareTray")
    can_toggle_chat: bool = Field(False, alias="canToggleChat")
    can_stop_sharing: bool = Field(False, alias="canStopSharing")
    can_pair: bool = Field(False, alias="canPair")


class MeetingState(BaseModel):
    """Current state of the meeting communicated by teams."""

    is_muted: bool = Field(False, alias="isMuted")
    is_hand_raised: bool = Field(False, alias="isHandRaised")
    is_in_meeting: bool = Field(False, alias="isInMeeting")
    is_recording_on: bool = Field(False, alias="isRecordingOn")
    is_background_blurred: bool = Field(False, alias="isBackgroundBlurred")
    is_sharing: bool = Field(False, alias="isSharing")
    has_unread_messages: bool = Field(False, alias="hasUnreadMessages")
    is_video_on: bool = Field(False, alias="isVideoOn")


class MeetingUpdate(BaseModel):
    """Update of the meeting state communicated by teams."""

    meeting_permissions: Optional[MeetingPermissions] = Field(
        None,
        alias="meetingPermissions",
    )
    meeting_state: Optional[MeetingState] = Field(None, alias="meetingState")


class ServerMessage(BaseModel):
    """Message received from the Teams WebSocket server."""

    request_id: Optional[int] = Field(None, alias="requestId")
    response: Optional[str] = None
    error_msg: Optional[str] = Field(None, alias="errorMsg")
    token_refresh: Optional[str] = Field(None, alias="tokenRefresh")
    meeting_update: Optional[MeetingUpdate] = Field(None, alias="meetingUpdate")


class ClientMessageParameterType(str, Enum):
    """Types of reactions for client messages."""

    ReactApplause = "applause"
    ReactLaugh = "laugh"
    ReactLike = "like"
    ReactLove = "love"
    ReactWow = "wow"
    ToggleUiChat = "chat"
    ToggleUiSharing = "sharing-tray"


class ClientMessageParameter(BaseModel):
    type_: ClientMessageParameterType = Field(..., serialization_alias="type")


class MeetingAction(str, Enum):
    NoneAction = "none"
    QueryMeetingState = "query-state"
    Mute = "mute"
    Unmute = "unmute"
    ToggleMute = "toggle-mute"
    HideVideo = "hide-video"
    ShowVideo = "show-video"
    ToggleVideo = "toggle-video"
    UnblurBackground = "unblur-background"
    BlurBackground = "blur-background"
    ToggleBlurBackground = "toggle-background-blur"
    LowerHand = "lower-hand"
    RaiseHand = "raise-hand"
    ToggleHand = "toggle-hand"
    LeaveCall = "leave-call"
    React = "send-reaction"
    ToggleUI = "toggle-ui"
    StopSharing = "stop-sharing"


class ClientMessage(BaseModel):
    """Message sent to the Teams WebSocket server."""

    action: MeetingAction
    parameters: Optional[ClientMessageParameter] = None
    request_id: int = Field(None, serialization_alias="requestId")

    _request_id_counter: ClassVar[int] = 0
    """Each message must have a unique request_id. The counter is incremented for each message automatically."""

    @classmethod
    def create(cls, *args, **kwargs) -> "ClientMessage":
        cls._request_id_counter += 1
        kwargs["request_id"] = cls._request_id_counter
        return cls(*args, **kwargs)
